/*
 * SPDX-FileCopyrightText: Stone Tickle <lattis@mochiro.moe>
 * SPDX-License-Identifier: GPL-3.0-only
 */

#ifndef MUON_FUNCTIONS_CONFIGURATION_DATA_H
#define MUON_FUNCTIONS_CONFIGURATION_DATA_H
#include "lang/func_lookup.h"

extern const struct func_impl impl_tbl_configuration_data[];
#endif
