/*
 * SPDX-FileCopyrightText: Stone Tickle <lattis@mochiro.moe>
 * SPDX-License-Identifier: GPL-3.0-only
 */
#ifndef MUON_LANG_LSP_H
#define MUON_LANG_LSP_H
#include <stdbool.h>
struct az_opts;
bool analyze_server(struct az_opts *opts);
#endif
