/*
 * SPDX-FileCopyrightText: Stone Tickle <lattis@mochiro.moe>
 * SPDX-License-Identifier: GPL-3.0-only
 */

#ifndef MUON_BACKEND_INTROSPECT_H
#define MUON_BACKEND_INTROSPECT_H
#include "lang/workspace.h"

bool introspect_write_all(struct workspace *wk);
#endif
