/*
 * SPDX-FileCopyrightText: Stone Tickle <lattis@mochiro.moe>
 * SPDX-License-Identifier: GPL-3.0-only
 */

#include "compat.h"

#include "ui.h"

bool have_ui = false;

bool
ui_main(void)
{
	return false;
}
