/*
 * SPDX-FileCopyrightText: Stone Tickle <lattis@mochiro.moe>
 * SPDX-License-Identifier: GPL-3.0-only
 */

#ifndef MUON_FUNCTIONS_KERNEL_CONFIGURE_FILE_H
#define MUON_FUNCTIONS_KERNEL_CONFIGURE_FILE_H
#include "lang/workspace.h"

bool func_configure_file(struct workspace *wk, obj _, obj *res);
#endif
