/*
 * SPDX-FileCopyrightText: Stone Tickle <lattis@mochiro.moe>
 * SPDX-License-Identifier: GPL-3.0-only
 */

body {
  max-width: 700px;
  margin: auto;
  padding: 10px;
  font-family: sans-serif;
}

@media (prefers-color-scheme: dark) {
  body {
      background-color: #282A36;
      color: #F8F8F2;
  }
  a {
    color: #8BE9FD;
  }

  small {
    color: #e8e8e2;
  }
}

nav div {
  padding: 10px;
  padding-left: 20px;
  display: block;
}

.logo {
  display: flex;
  flex-direction: row;
  align-items: center;
  font-size: 40px;
}

.logo small {
  font-size: 10px;
}

.logo_text {
  display: flex;
  flex-direction: column;
  padding: 10px;
}

.loader {
  width: 18px;
  height: 18px;
  border: 5px solid #50FA7B;
  border-bottom-color: transparent;
  border-radius: 50%;
  display: inline-block;
  box-sizing: border-box;
  animation: rotation 1s linear infinite;
}

@keyframes rotation {
  0% {
    transform: rotate(0deg);
  }
  100% {
    transform: rotate(360deg);
  }
}

.version-item {
  display: inline-flex;
  flex-direction: row;
  justify-content: flex-start;
}

summary {
  margin-bottom: 10px;
}

summary:hover {
  background-color: #21222C;
}

.version-tag {
  margin-left: 3px;
  font-size: 10px;
  border-radius: 40% / 100%;
  padding: 1px;
  min-width: 40px;
  display: flex;
  flex-direction: column;
  align-items: center;
  justify-content: center;
}

.version-tag-dev {
  color: #282A36;
  background-color: #FFB86C;
}

.version-tag-latest {
  color: #282A36;
  background-color: #50FA7B;
}
